/* 
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or (at
   your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   Copyright (c) Alexey Mahotkin <alexm@hsys.msk.ru> 2001

   "cvs acl" subcommand manages ACL based on working directory

*/


#include "config.h"

#include "cvs.h"
#include "acl.h"
#include "repos.h"

#include "acl-admins.h"
#include "branch-acl.h"
#include "dir-acl.h"
#include "stringbuf.h"

#include <stdlib.h>

#ifdef SERVER_SUPPORT

int
grant_file_permission (char * const dir, char * const branch, char * const filename, 
		       branch_permission perm, char *perm_string)
{
    struct DIR_ACL *dir_acl;
    struct BRANCH_ACL *branch_acl;
    struct BRANCH_ACE *branch_ace;

    char *xbranch = branch;

    if ((xbranch == NULL) || (xbranch[0] == '\0'))
	xbranch = "trunk";

    if (!server_active)
	return 1;

    dir_acl = get_dir_acl(dir, "trunk");
    if (dir_acl == NULL)
	error(1, 0, "Error getting directory ACL for %s: %s", dir, strerror(errno));

    branch_acl = get_branch_acl(dir_acl, xbranch);
    if (branch_acl == NULL)
	error(1, 0, "Error getting branch ACL for %s: %s", dir, strerror(errno));

    branch_ace = get_file_on_branch_ace(branch_acl, filename);
    if (branch_ace == NULL)
	error(1, 0, "Error getting file ACE for %s: %s", filename, strerror(errno));

    if (!branch_permission_granted(branch_ace, CVS_Username, perm)) {

	error(0, 0, "%s permission for %s denied for user %s", 
	      perm_string, filename, CVS_Username);

	return 0;
    }

    return 1;
}

int
grant_dir_permission (char * const dir, char * const filename,
		      dir_permission perm, char *perm_string)
{
    DIR_ACL *dir_acl;

    if (!server_active)
	return 1;

    dir_acl = get_dir_acl(dir, "trunk");
    if (dir_acl == NULL)
	error(1, 0, "Error getting directory ACL for %s: %s", dir, strerror(errno));

    if (dir_permission_granted(dir_acl, CVS_Username, perm) != perm) {
	error(0, 0, "%s permission for %s denied for user %s",
	      perm_string, filename, CVS_Username);
	return 0;
    }

    return 1;
}

#endif /* SERVER_SUPPORT */

static char *
full_relative_dirname (char *dirname)
{
    char *rel_dir;
    char *p;
    int len;

    rel_dir = Name_Repository(dirname, NULL);
    len = strlen(rel_dir);

    p = realloc(rel_dir, len + 1);
    if (p == NULL)
	return NULL;

    rel_dir[len++] = '/';
    rel_dir[len] = '\0';

    return Short_Repository(rel_dir);
}


static char *
full_relative_filename (char *filename)
{
    char *p = last_component(filename);
    char *rel_dir;
    if (p == filename)
	rel_dir = Name_Repository(".", NULL);
    else {
	char *dirname = filename;
	filename = p;
	p--;
	*p = '\0';
	rel_dir = Name_Repository(dirname, NULL);
    }

    p = realloc(rel_dir, strlen(rel_dir) + 1 + strlen(filename) + 1);
    if (p == NULL) {
	free(rel_dir);
	return NULL;
    }

    strcat(p, "/");
    strcat(p, filename);

    return Short_Repository(p);
}


static const char *const acl_usage[] = 
{
    "Usage: %s %s [-r BRANCH] grant USER permissions files...\n", 
    "\n",
    "\t\tor:\n",
    "\n",
    "\t%s %s [-r BRANCH] ls USER files...\n",
    "\n",
    NULL
};


struct acl_common_data {
    char* subcommand;
    char *username;
    char *branch;
    char *perm_string;
    char **filenames;
    int filescount;
};

static struct acl_common_data static_common_data;

static int
set_acl_for_file (char *relative_name, char *branch, 
		  char *perm_string, char *username)
{
    struct stringbuf *fulldir;
    struct DIR_ACL *dir_acl;
    struct BRANCH_ACL *branch_acl;
    struct BRANCH_ACE *file_ace;
    branch_permission branch_perm;
    char *p;
    char *filename;

    fulldir = new_stringbuf(current_parsed_root->directory);
    if (fulldir == NULL)
	return -1;

    catc_stringbuf(fulldir, '/');
    cat_stringbuf(fulldir, relative_name);
    p = last_component(fulldir->buf);
    filename = p;

    p--;
    *p = '\0';

    dir_acl = get_dir_acl(fulldir->buf, branch);
    if (dir_acl == NULL)
	return -1;

    branch_acl = get_branch_acl(dir_acl, branch);
    if (branch_acl == NULL)
	return -1;

    file_ace = get_file_on_branch_ace(branch_acl, filename);
    if (file_ace == NULL)
	return -1;

    branch_perm = parse_branch_permissions(perm_string);
    if (branch_perm == branch_perm_invalid)
	error(1, 0, "Invalid branch permissions '%s'", perm_string);
    set_branch_ace(file_ace, username, branch_perm);

    if (!store_branch_acls(dir_acl))
	error(1, 0, "Error storing branch ACLs: %s", strerror(errno));

    printf("Setting '%s' permissions for '%s' on file %s\n", 
	   perm_string, username, relative_name);

    return 0;
}


static int
set_acl_for_dir (char *relative_name, char *branch, 
		 char *perm_string, char *username)
{
    dir_permission perm;
    struct stringbuf *fulldir;
    DIR_ACL *dir_acl;
    int default_len = strlen("default:");

    fulldir = new_stringbuf(current_parsed_root->directory);
    if (fulldir == NULL)
	return -1;

    catc_stringbuf(fulldir, '/');
    cat_stringbuf(fulldir, relative_name);

    dir_acl = get_dir_acl(fulldir->buf, branch);
    if (dir_acl == NULL)
	return -1;

    if (strncmp(perm_string, "default:", default_len) == 0) {
	/* set default file permissions */
	perm = parse_branch_permissions(perm_string + default_len);
	if (perm == branch_perm_invalid)
	    error(1, 0, "Invalid default file permissions '%s'", perm_string);

	if (!set_default_file_acl(dir_acl, branch, username, perm))
	    error(1, 0, "Error setting default file permission %s on directory %s",
		  perm_string, relative_name);

	printf("Setting default file permissions '%s' for '%s' on directory %s\n", 
	       perm_string, username, relative_name);

    } else {
	/* set directory permission */
	perm = parse_dir_permissions(perm_string);
	if (perm == dir_perm_invalid)
	    error(1, 0, "Invalid directory permissions '%s'", perm_string);

	if (set_dir_ace(dir_acl, username, perm) != perm)
	    error(1, 0, "Error setting %s permissions on directory %s",
		  perm_string, relative_name);

	printf("Setting '%s' permissions for '%s' on directory %s\n", 
	       perm_string, username, relative_name);

    }

    if (store_dir_acl(dir_acl) != 0)
	error(1, 0, "Error storing permissions on directory %s",
	      relative_name);

    return 0;
}

enum processing_what { processing_started, processing_files, processing_dirs };

static void
common_acl (int argc, char **argv, struct acl_common_data *common_data)
{
    int c;
    char *branch = "trunk";
    char *subcommand;
    char *username;
    char *perm_string;
    enum processing_what processing = processing_started;

    if (argc == -1)
	usage(acl_usage);

    optind = 0;
    while ((c = getopt(argc, argv, "r:")) != -1) {
	switch (c) {
	case 'r':
	    branch = xstrdup(optarg);
	    break;

	default:
	    error(1, 0, "invalid option -%c\n", c);
	    usage(acl_usage);
	    break;
	}
    }
    argc -= optind;
    argv += optind;

    common_data->branch = branch;

    /* take the acl-subcommand */
    subcommand = *argv++;
    argc--;
    if (subcommand == NULL)
	usage(acl_usage);

    /* learn the username */
    username = *argv++;
    argc--;
    if (username == NULL)
	usage(acl_usage);

    /* learn the permissions */
    perm_string = *argv++;
    argc--;
    if (perm_string == NULL)
	usage(acl_usage);

    common_data->filenames = argv;
    common_data->filescount = argc;

    /* process each file or directory */
    for (; *argv; argv++) {
	char *filename = *argv;
	int namelen;

	namelen = strlen(filename);

	if ((*filename == '\0') || (filename[namelen - 1] == '/')) {
	    if ((processing != processing_started) &&
		(processing != processing_dirs))
	    {
		error(1, 0, "Directory %s appeared while processing files", 
		      filename);
	    }

	    processing = processing_dirs;
	} else {
	    if ((processing != processing_started) &&
		(processing != processing_files))
	    {
		error(1, 0, "File %s appeared while processing directories", 
		      filename);
	    }

	    processing = processing_files;
	}
#ifdef SERVER_SUPPORT
	if (server_active) {
	    if (strcmp(subcommand, "grant") == 0) {
		if (processing == processing_files) {
		    if (set_acl_for_file(filename, branch, perm_string, username) != 0)
			error(0, 0, "Error setting ACLs for file %s: %s",
			      filename, strerror(errno));
		} else if (processing == processing_dirs) {
		    if (set_acl_for_dir(filename, branch, perm_string, username) != 0)
			error(0, 0, "Error setting ACLs for directory %s: %s",
			      filename, strerror(errno));
		} else
		    error(1, 0, "Unknown type of argument %s", filename);
	    } else
		usage(acl_usage);
	} else {
#endif
	    if (processing == processing_files) {
		char *new_filename = full_relative_filename(filename);
		if (new_filename == NULL)
		    error(1, 0, "Out of memory");
		
		*argv = new_filename;
	    } else if (processing == processing_dirs) {
		char *new_dirname = full_relative_dirname(filename);
		if (new_dirname == NULL)
		    error(1, 0, "Out of memory");

		*argv = new_dirname;
	    } else
		error(1, 0, "Unknown type of argument %s", filename);
#ifdef SERVER_SUPPORT
	}
#endif

    }
    common_data->subcommand = subcommand;
    common_data->username = username;
    common_data->perm_string = perm_string;
}


#ifdef CLIENT_SUPPORT
static int
client_side_acl (struct acl_common_data *common_data)
{
    start_server();
    
    if (!supported_request ("acl"))
	error (1, 0, "server does not support acl");

    option_with_arg ("-r", common_data->branch);

    send_arg(common_data->subcommand);
    send_arg(common_data->username);
    send_arg(common_data->perm_string);

    send_file_names (common_data->filescount, common_data->filenames,
		     SEND_EXPAND_WILD);
    send_to_server ("acl\012", 0);

    return get_responses_and_close();
}
#endif


int
acl (int argc, char *argv[])
{
#ifdef SERVER_SUPPORT
    if (server_active) {
	if (!is_acl_admin(CVS_Username))
	    error(1, 0, "%s is not allowed to manage ACLs", CVS_Username);
    }
#endif

    common_acl(argc, argv, &static_common_data);

#ifdef CLIENT_SUPPORT
    if (current_parsed_root->isremote)
	return client_side_acl(&static_common_data);
#endif

    return 0;
}
