#include "config.h"

#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif

#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <errno.h>
#include <sys/types.h>
#include <stdio.h>

#include "network-client.h"
#include "socket-client.h"
#include "http_proxy-client.h"
#include "error.h"

int
ncli_need_http_proxy (NCLI ncli, char *hostname, int port)
{
	char *proxy = getenv(HTTP_PROXY_VAR);
	if (proxy == NULL || *proxy == '\0')
		return 0;
	/* more checks regarding HTTP_NOPROXY_VAR ($CVS_NOPROXY) */
	return 1;
}

#define SCHEME_SUFFIX "://"

static int
http_proxy_get_info(NCLI ncli, char *m_proxy,
					char **host, unsigned short *port,
					char **user, char **password)
{
	{
		size_t len = strlen(m_proxy);
		if (m_proxy[len-1] == '/')
			m_proxy[len-1] = '\0';
	}

	{
		char *scheme = strstr(m_proxy, SCHEME_SUFFIX);
		if (scheme != NULL)
			(*user) = scheme + (sizeof(SCHEME_SUFFIX) - 1);
		else
			(*user) = m_proxy;
	}
	
	if (((*host) = strchr((*user), '@'))!=NULL) {
		(**host) = '\0';
		(*host)++;
	} else {
		(*host) = (*user);
		(*user) = NULL;
	}
	
	if ((*user) != NULL && ((*password) = strchr((*user), ':'))!=NULL) {
		(**password) = '\0';
		(*password)++;
	} else
		(*password) = NULL;

	{
		char *port_ptr = strchr((*host), ':');
		if (port_ptr != NULL) {
			char *endp;
			unsigned long ul_port;
			(*port_ptr) = '\0';
			port_ptr++;
			ul_port = strtoul(port_ptr, &endp, 10);
			if ((*endp) != '\0' || ul_port == 0 || ul_port > USHRT_MAX) {
				error(0, 0, "Invalid port specification in $" HTTP_PROXY_VAR " '%s'",
						getenv(HTTP_PROXY_VAR));
				return 0;
			}
			(*port) = (unsigned short)ul_port;
		} else
			(*port) = 3128;
	}
	return 1;
}

/* table is borrowed from mutt */
static char B64Chars[64] = {
	'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M',
	'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z',
	'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm',
	'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z',
	'0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '+', '/'
};

static int
b64_encode(const unsigned char *s, FILE *f) {
	int rembits = 2;
	int remainder = 0;
	while (*s) {
		if (putc(B64Chars[remainder | ((*s) >> rembits)], f)==EOF)
			return 0;
		remainder = (((*s) << (6 - rembits)) & 0x3F) ;
		rembits += 2;
		s++;
		if (rembits == 8) {
			if (putc(B64Chars[remainder], f)==EOF)
				return 0;
			remainder = 0;
			rembits = 2;
		}
	}
	if (rembits) {
		if (putc(B64Chars[remainder], f)==EOF)
			return 0;
	}
	return 1;
}

#ifdef CR
# undef CR
#endif
#define CR '\015'

#ifdef LF
# undef LF
#endif
#define LF '\012'

#ifdef CRLF
# undef CRLF
#endif

#define CRLF "\015\012"

static int
http_proxy_negotiate(NCLI ncli, int sockfd,
					 const char *host, const unsigned short port,
					 const char *user, const char *password)
{
	FILE *f;
	sockfd = dup(sockfd);
	if (sockfd == -1) {
		error(0, errno, "Unable to create socket duplicate");
		return NCLI_FAILURE;
	}
	if ((f = fdopen(sockfd,"a+")) == NULL) {
		error(0,errno, "Unable to create FILE");
		close(sockfd);
		return NCLI_FAILURE;
	}
	setvbuf(stdin,(char*)NULL,_IOLBF,0);
	
	if (fprintf(f,"CONNECT %s:%hu HTTP/1.1" CRLF "Host: %s: %hu" CRLF,
				host, port, host, port)<=0) {
		error(0,errno,"A write error while negotiating proxy connection");
		fclose(f);
		return NCLI_FAILURE;
	}
	if (user && password) {
		/* support only for basic auth */
		if (fprintf(f,"Authorization: Basic ") <= 0) {
			error(0, errno, "A write error while negotiating proxy connection");
			fclose(f);
			return NCLI_FAILURE;
		}
		{
			/* maybe we should use cheat mode? */
			int result;
			char *buf = malloc(strlen(user) + strlen(password)+2);
			if (buf == NULL) {
				error(0,ENOMEM, "Authorization info error");
				fclose(f);
				return NCLI_FAILURE;
			}
			/* we made it big enough already */
			sprintf(buf,"%s:%s",user,password);
			result = b64_encode(buf,f);
			free(buf);
			if (!result) {
				error(0,errno, "Sending proxy authorization info error");
				fclose(f);
				return NCLI_FAILURE;
			}
			if (fprintf(f,CRLF) <= 0) {
				error(0,errno, "A proxy negotiation error");
				fclose(f);
				return NCLI_FAILURE;
			}
		}
	}
	if (fprintf(f,CRLF) <= 0) {
		error(0,errno, "A proxy negotiation error");
		fclose(f);
		return NCLI_FAILURE;
	}
	{
		char buf[4];
		if (fscanf(f,"HTTP/%*[0-9].%*[0-9] %3[0-9] %*s" CRLF, buf) < 1) {
			error(0,0,"Invalid proxy response format");
			fclose(f);
			return NCLI_FAILURE;
		}
		
		if (strncmp(buf,"200",3)!=0) {
			buf[3] = '\0';
			error(0,0,"proxy access failure: code %s\n", buf);
			fclose(f);
			return NCLI_FAILURE;
		}
	}
	{
		/* well, now we may skip everything else up to an empty string
		 * (two CRLFs) */
		int errcode = 0;
		char *errstr = NULL;
		int c;
		while (1) {
			unsigned int len = 0;
			while (1) {
				if ((c=fgetc(f))==EOF) {
					errstr = "proxy response read error";
					errcode = errno;
					break;
				}
				if ((unsigned char)c == LF)
					break;
				else if ((unsigned char)c != CR)
					len++;
			}
			if (len == 0 || errstr != NULL)
				break;
		}
		fclose(f);
		if (errstr != NULL) {
			error(0,errcode,"%s",errstr);
			return NCLI_FAILURE;
		}
	}
	return NCLI_SUCCESS;
}

int ncli_connect_http_proxy(NCLI ncli, char *shost, int sport)
{
	char *m_proxy;
	char *host = NULL;
	unsigned short port;
	char *user = NULL;
	char *password = NULL;
	int sock;
	int result;
	
	if ((m_proxy = getenv(HTTP_PROXY_VAR)) == NULL)
		return ncli->error(ncli,"$" HTTP_PROXY_VAR " is not specified");

	if ((m_proxy = strdup(m_proxy)) == NULL)
		return ncli->error(ncli, "Not enough memory");

	if (!http_proxy_get_info(ncli,m_proxy, &host,&port,&user,&password)) {
		free(m_proxy);
		return ncli->error(ncli,"Invalid proxy specification");
	}
	
	sock = ncli_connect_socket_direct(ncli,host,port);
	
	if (sock < 0) {
		free(m_proxy);
		return ncli->error(ncli,"Unable to connect to proxy");
	}

	/* perform the proxy protocol negotiations */
	result = http_proxy_negotiate(ncli,sock,shost,sport,user,password);
	free(m_proxy);

	if (result != NCLI_SUCCESS)
		return ncli->error(ncli,"Unable to connect to proxy");
    return sock;
}
   
