/* 
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or (at
   your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   Copyright (c) Alexey Mahotkin <alexm@hsys.msk.ru> 1999-2000

   lockpasswd() and unlockpasswd() prevent simultaneous changing of
   CVSROOT/cvspasswd by several users.

*/

#define _BSD_SOURCE 1

#include "config.h"

#include <stdlib.h>
#include <string.h>

#ifdef HAVE_NETDB_H
#include <netdb.h>
#endif

#include "cvs.h"

#include "lockpasswd.h"

static char*
lockfilename PROTO((char *repos))
{
    char *lockname;
    int len = strlen(repos) + 1;
    len += strlen(CVSROOTADM "/" CVSROOTADM_CVSPASSWD) + 1;
    len += strlen(".lock") + 1;
    
    lockname = malloc(len);
    if (!lockname)
    {
	fprintf(stderr, "virtual memory exhausted\n");
	exit(EXIT_FAILURE);
    }

    snprintf(lockname, len, "%s/%s/%s.lock",
	     repos, CVSROOTADM, CVSROOTADM_CVSPASSWD);

    return(lockname);
}

static char *global_lockname;

static void internal_unlock(char *lockname);

void atexit_unlockpasswd PROTO((void));

void 
atexit_unlockpasswd(void)
{
    if (global_lockname)
	internal_unlock(global_lockname);
}

void
lockpasswd PROTO((char *repos))
{
    struct stat buf;
    int fd0;
    char *hostnam;
    int len;
    char *lockname0, *lockname1;
    int lockmade = 0;

    hostnam = malloc(MAXHOSTNAMELEN + 1);
    if (!hostnam) 
    {
	fprintf(stderr, "virtual memory exhausted\n");
	exit(EXIT_FAILURE);
    }
    gethostname(hostnam, MAXHOSTNAMELEN);

    len = strlen(repos) + 1;
    len += strlen(CVSROOTADM "/" CVSROOTADM_CVSPASSWD) + 1;
    len += strlen(hostnam) + 1;
    len += 10 + 1; /* strlen(2^32) */
    len += strlen(".lock") + 1;

    lockname0 = malloc(len);
    if (!lockname0)
    {
	fprintf(stderr, "virtual memory exhausted\n");
	exit(EXIT_FAILURE);
    }
    snprintf(lockname0, len, "%s/%s/%s.%s-%d.lock", 
	     repos, CVSROOTADM, CVSROOTADM_CVSPASSWD, hostnam, getpid());

    free(hostnam);

    /* printf("Lockfile is named %s\n", lockname0); */

    fd0 = CVS_OPEN(lockname0, O_CREAT | O_EXCL, 0600);
    if (fd0 == -1)
    {
	fprintf(stderr, "Error creating lock-file %s: %s\n", 
		lockname0, strerror(errno));
	exit(EXIT_FAILURE);
    }

    lockname1 = lockfilename(repos);

    /* printf("New lock-file name is %s\n", lockname1); */

    while (!lockmade) {

	link(lockname0, lockname1);
	    
	if (fstat(fd0, &buf) == -1)
	{
	    fprintf(stderr, "Cannot stat() lock-file %s: %s\n", 
		    lockname0, strerror(errno));
	    exit(EXIT_FAILURE);
	}

	lockmade = (buf.st_nlink == 2);
	
	if (lockmade)
	{
	    if (unlink(lockname0) == -1)
	    {
		fprintf(stderr, "Cannot unlink() lock-file %s: %s\n", 
			lockname0, strerror(errno));
		exit(EXIT_FAILURE);
	    }
	}
	else
	{
	    printf("Waiting for lock to disappear\n");
	    sleep(3);
	}
    }

    if (close(fd0) == -1)
    {
	fprintf(stderr, "Error closing lock-file: %s", strerror(errno));
	exit(EXIT_FAILURE);
    }

    /* if (lockmade)
       printf("Locked!\n"); */

    global_lockname = lockname1;
    atexit(atexit_unlockpasswd);

    free(lockname0);
}

void
unlockpasswd PROTO((char *repos))
{
    char *lockname = lockfilename(repos);
    internal_unlock(lockname);
    global_lockname = NULL;
}

static void
internal_unlock (char *lockname)
{
    if (unlink(lockname) == -1)
    {
	fprintf(stderr, "Cannot unlink() lock-file %s: %s\n", 
		lockname, strerror(errno));
	exit(EXIT_FAILURE);
    }
}
