/* 
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or (at
   your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   Copyright (c) Alexey Mahotkin <alexm@hsys.msk.ru> 2001

   "ls" subcommand gives a listing of specified directory in
   repository

*/

#include "config.h"

#include "ls.h"

#include "stringbuf.h"

#include "cvs.h"
#include "acl.h"
#include "recurse.h"

#include <stdlib.h>

static const char *const ls_usage[] =
{
    "Usage: %s %s subdirectories...\n",
    "\t-R\tRecursively list subdirectories.\n",
    "\t-r TAG\tUse specified tag.\n",
    "\t-D DATE\tUse specified date.\n",
    "\t-f\tForce tag match.\n",
    NULL
};

static int local_only = 1;
static int force_tag_match = 0;
static char *tagname = NULL;
static char *date = NULL;
static int deep = 0;
static struct DIR_ACL *dir_acl = NULL;


/* ARGSUSED */
static Dtype
ls_dirproc (callerdat, dir, repository, update_dir, entries)
    void *callerdat;
    char *dir;
    char *repository;
    char *update_dir;
    List *entries;
{
    /* Should we go this way? AM */
#ifdef SERVER_SUPPORT
    int perm;
    char *xbranch = (tagname) ? tagname : "trunk";
    if ((dir_acl = get_dir_acl(repository,xbranch)) == NULL)
	error(1, 0, "Error getting directory ACL for %s: %s", repository, strerror(errno));
    perm = dir_permission_granted(dir_acl,CVS_Username,dir_perm_all);
#else
    int perm = dir_perm_all;
#endif

    if (deep == 0 && !(perm & dir_perm_access)) {
	/* prevent unprivileged user from checking if a particular
	 * directory really exists
	 */
	error(1, 0, "could not read RCS file for %s",update_dir);
	/* just in case */
	return R_SKIP_ALL;
    }
    deep++;
	
    cvs_output("D/", 2);
    cvs_output(update_dir,0);
    cvs_output("//",2);
    if (perm & dir_perm_access)
	cvs_output("r",1);
    else
	cvs_output("-",1);
    if (perm & dir_perm_modify) 
	cvs_output("w",1);
    else
	cvs_output("-",1);
    cvs_output("\n", 1);

#ifdef SERVER_SUPPORT
    if (!(perm & dir_perm_access))
	return R_SKIP_ALL;
#endif
    if (local_only && deep > 1)
	return R_SKIP_ALL;
    return R_PROCESS;
}

static int
ls_dirleaveproc (void *callerdat, char *dir,
		 int err, char *update_dir,
		 List *entries)
{
    if (deep > 0)
	deep--;
    else
	cvs_outerr("Invalid recursion",0);

    return 0;
}


static int
ls_fileproc (void *callerdat, struct file_info *finfo)
{
    char *rev;
    /* int is_simple_tag = 1; */
    Node *rcsversnode;
    RCSVers *rcsvers;
    time_t revtime;
    struct BRANCH_ACL *branch_acl;
    struct BRANCH_ACE *branch_ace;
    char timestamp[32];
    int ts_len;
    int single_file;
    char *xbranch;
    int perm;
    Vers_TS *ts;

#ifdef SERVER_SUPPORT
    if (dir_acl == NULL) {
	xbranch = (tagname) ? tagname : "trunk";
	if ((dir_acl = get_dir_acl(finfo->repository,xbranch)) == NULL)
	    error(1, 0, "Error getting directory ACL for %s: %s", finfo->repository, strerror(errno));
	single_file = 1;
    }
#endif

    ts = Version_TS(finfo, NULL, tagname, date, force_tag_match, 0);

    cvs_output("/", 1);
    cvs_output(finfo->file, 0);
    cvs_output("/", 1);

    rev = ts->vn_rcs;
    if (tagname) {
	rev = RCS_gettag(ts->srcfile, tagname, force_tag_match, NULL);
	if (rev == NULL)
	    error(1, 0, "Unknown tag '%s'", tagname);
    }
    cvs_output(rev, 0);
    cvs_output("/", 1);

    RCS_fully_parse(ts->srcfile);
    rcsversnode = findnode(ts->srcfile->versions, ts->vn_rcs);
    rcsvers = (RCSVers *) rcsversnode->data;
    if (rcsvers == NULL)
	error(1, 0, "Internal error: cannot find revision '%s' of %s",
	      ts->vn_rcs, finfo->fullname);

    cvs_output(rcsvers->author, 0);
    cvs_output("/", 1);

    revtime = RCS_getrevtime(ts->srcfile, ts->vn_rcs, 0, 0);
    snprintf(timestamp, sizeof(timestamp), "%lu", revtime);
    cvs_output(timestamp, 0);
    cvs_output("/", 1);
#ifdef SERVER_SUPPORT
    branch_acl = get_branch_acl(dir_acl,xbranch);
    if (branch_acl == NULL)
	error(1, 0, "Error getting branch ACL for %s: %s", finfo->repository, strerror(errno));

    branch_ace = get_file_on_branch_ace(branch_acl, finfo->file);
    if (branch_ace == NULL)
	error(1, 0, "Error getting file ACE for %s: %s", finfo->file, strerror(errno));

    perm = branch_permission_granted(branch_ace, CVS_Username, branch_perm_all);
#else
    perm = branch_perm_all;
#endif
   
    if (perm & branch_perm_checkout)
	cvs_output("r",1);
    else
	cvs_output("-",1);
    if (perm & branch_perm_checkin) 
	cvs_output("w",1);
    else
	cvs_output("-",1);

    cvs_output("/",1);

    cvs_output("\n", 1);

    return 0;
}



static int 
ls_proc (int argc, char **argv, char *xwhere,
	 char *mwhere, char *mfile, int shorten,
	 int local, char *mname, char *msg)
{
    int err = 0;
    int which = W_REPOS | W_ATTIC;
    struct stringbuf *repository = NULL;

    repository = new_stringbuf(current_parsed_root->directory);

    if (CVS_CHDIR(repository->buf) < 0) {
	err = 1;
	goto out;
    }
	
    deep = 0;
    local = 0; /* we always use recursion mode, 'cause we anyway need
		  subdirectores listing
	       */
    err = start_recursion (ls_fileproc, (FILESDONEPROC) NULL, ls_dirproc,
			   ls_dirleaveproc, (void *) NULL,
			   argc - 1, argv + 1, local, which, 0, 1,
			   "", 1);

 out:
    free_stringbuf(repository);
    return err;
}

static int
server_side_ls (int argc, char **argv)
{
    int err = 0;
    int i;

    for (i = 0; i < argc; i++) {
	err += ls_proc (argc + 1, argv - 1, (char *) NULL,
			(char *) NULL, (char *) NULL, 0, local_only, (char *) NULL,
			(char *) NULL);
    }
    return err;
}

int
ls (int argc, char **argv)
{
    int i;
    int c;

    local_only = 1; /* Set the default */
    optind = 0;
    while ((c = getopt (argc, argv, "D:fRr:")) != -1)
    {
	switch (c)
	{
	case 'D':
	    date = xstrdup(optarg);
	    break;

	case 'f':
	    force_tag_match = 1;
	    break;
	    
	case 'R':
	    local_only = 0;
	    break;

	case 'r':
	    tagname = xstrdup(optarg);
	    break;

	default:
	    usage(ls_usage);
	    break;
	}

    }
    argc -= optind;
    argv += optind;
    if (argc < 1 || !(*argv[0]))
	usage (ls_usage);
	
#ifdef CLIENT_SUPPORT
    if (current_parsed_root->isremote) {

	start_server();

	ign_setup();

	if (!local_only)
	    send_to_server ("Argument -R\012", 0);

	if (force_tag_match)
	    send_to_server ("Argument -f\012", 0);

	if (tagname != NULL)
	    option_with_arg ("-r", tagname);

	if (date != NULL)
	    option_with_arg ("-D", date);

	for (i = 0; i < argc; i++)
	    send_arg (argv[i]);

	send_to_server("ls\012", 0);

	return get_responses_and_close();
    }
#endif /* CLIENT_SUPPORT */

    return server_side_ls(argc, argv);    
}
