/* 
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or (at
   your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   Copyright (c) Alexey Mahotkin <alexm@hsys.msk.ru> 1999-2000

   "passwd" manages CVSROOT/cvspasswd file

*/

#define _BSD_SOURCE 1

#include "config.h"

#include "cvs.h"
#include "getline.h"

#include "passwd.h"

#include "cvs-server.h"
#include "cvspasswd-file.h"
#include "gensalt.h"
#include "lockpasswd.h"
#include "scramble.h"

#include <assert.h>
#ifdef HAVE_CRYPT_H
#include <crypt.h>
#endif

#ifdef SERVER_SUPPORT

static const char *const passwd_usage[] = 
{
    "Usage: %s %s [-a|-e|-d|-c|-x|-p] username [-i info] [-m e-mail@address]\n",
    "\t-a\tAdd new user (account disabled by default)\n",
    "\t-e\tEnable user account\n",
    "\t-d\tDisable user account\n",
    "\t-c\tChange user account properties\n",
    "\t-x\tRemove user account\n",
    "\t-p\tChange password (ordinary users may do this for themselves)\n",
    "\n",
    "Options:\n",
    "\t-u username[.group]\tChange system user and group\n",
    "\t-m e-mail@address\tChange user's e-mail address\n",
    "\t-i info\t\t\tChange user's information\n",
    "\t-t time\t\t\tChange user's account expiration time\n",
    "\t-p\t\t\tChange user's password\n",
    NULL
};

enum passwd_action {action_none,
		    action_add, 
		    action_change,
		    action_passwd,
		    action_remove, 
		    action_enable,
		    action_disable};

struct CVS_user {
    char *name;
    char *password;
    char *old_password;
    char *system_user;
    char *email;
    char *gecos;
    char *status;
};

static int
passwd_add_user PROTO((struct CVS_user user, FILE *cvspasswd, FILE *tmpcvspasswd));

static int
passwd_add_user(user, cvspasswd, tmpcvspasswd)
    struct CVS_user user;
    FILE *cvspasswd;
    FILE *tmpcvspasswd;
{
    char *salt;

    if (cvspasswd != NULL)
	find_user (cvspasswd, tmpcvspasswd, user.name, find_action_error);
    fprintf(tmpcvspasswd, "%s:", user.name);
    salt = gensalt ();
    if (!salt)
	error(1, 0, "Out of memory");

    fprintf(tmpcvspasswd, "%s:", crypt(user.password, salt));
    free(salt);

    fprintf(tmpcvspasswd, "%s:", user.system_user);

    if (user.email)
	fputs(user.email, tmpcvspasswd);
    fputc(':', tmpcvspasswd);

    if (user.gecos)
	fputs(user.gecos, tmpcvspasswd);
    fputc(':', tmpcvspasswd);
    
    fputs(user.status, tmpcvspasswd);

    fputc('\n', tmpcvspasswd);
    return 1;
}


static int
passwd_edit_user PROTO((struct CVS_user user, FILE *cvspasswd, FILE *tmpcvspasswd));

static int
passwd_edit_user(user, cvspasswd, tmpcvspasswd)
    struct CVS_user user;
    FILE *cvspasswd;
    FILE *tmpcvspasswd;
{
    char *user_line = find_user (cvspasswd, tmpcvspasswd, user.name, find_action_return);
    char *field_user, *field_password, *field_system_user;
    char *field_email, *field_gecos, *field_status;

    if (!user_line)
	error(1, 0, "user %s does not exist\n", user.name);

#define EXTRACT_NEXT_FIELD(field1, field2)			 \
field2 = strchr(field1, ':');					 \
if (!field2)							 \
    error(1, 0, "broken cvspasswd line for user %s", user.name); \
								 \
*field2++ = '\0';						 \

    field_user = user_line;
    EXTRACT_NEXT_FIELD(field_user, field_password);
    EXTRACT_NEXT_FIELD(field_password, field_system_user);
    EXTRACT_NEXT_FIELD(field_system_user, field_email);
    EXTRACT_NEXT_FIELD(field_email, field_gecos);
    EXTRACT_NEXT_FIELD(field_gecos, field_status);

    fputs(field_user, tmpcvspasswd);
    fputc(':', tmpcvspasswd);

    if (user.password) {
	char *salt = gensalt();
	if (!salt)
	    error(1, 0, "Out of memory!");

	fputs(crypt(user.password, salt), tmpcvspasswd);
	free(salt);
    } else
	fputs(field_password, tmpcvspasswd);
    fputc(':', tmpcvspasswd);

    fputs(field_system_user, tmpcvspasswd);
    fputc(':', tmpcvspasswd);

    if (user.email)
	fputs(user.email, tmpcvspasswd);
    else
	fputs(field_email, tmpcvspasswd);
    fputc(':', tmpcvspasswd);

    if (user.gecos)
	fputs(user.gecos, tmpcvspasswd);
    else
	fputs(field_gecos, tmpcvspasswd);
    fputc(':', tmpcvspasswd);

    if (user.status)
	fputs(user.status, tmpcvspasswd);
    else
	fputs(field_status, tmpcvspasswd);
    fputc('\n', tmpcvspasswd);
    
    find_user (cvspasswd, tmpcvspasswd, user.name, find_action_do_none);
    return 1;
}


static int
passwd_passwd PROTO((struct CVS_user user));

static int
passwd_passwd(user)
    struct CVS_user user;
{
    char *cvspasswd_argv[] = {"", NULL};
    int pi[2];
    int pid;
    char *cvspasswd = getenv("CVSPASSWD");
    if (!cvspasswd)
	error(1, 0, "Environment variable CVSPASSWD is not defined");

    cvspasswd_argv[0] = cvspasswd;

    if (!create_pipe (pi))
	error(1, 0, "cannot create a pipe: %s\n", strerror(errno));

    pid = fork ();

    if (pid == -1)
	error(1, 0, "cannot fork(): %s\n", strerror(errno));

    if (pid == 0)
    {
        close(pi[1]);
        execvp(cvspasswd, cvspasswd_argv);
	error(1, 0, "cannot exec(%s): %s\n", cvspasswd, strerror(errno));
    }

    close(pi[0]);
    write(pi[1], user.old_password, strlen(user.old_password) + 1);
    write(pi[1], user.password, strlen(user.password) + 1);
    close(pi[1]);
    memset(user.password, '\0', strlen(user.password));
    memset(user.old_password, '\0', strlen(user.old_password));
   
    return 1;
}

static int
passwd_remove_user PROTO((struct CVS_user user, FILE *cvspasswd, FILE *tmpcvspasswd));

static int
passwd_remove_user(user, cvspasswd, tmpcvspasswd)
    struct CVS_user user;
    FILE *cvspasswd;
    FILE *tmpcvspasswd;
{
    int result;
    result = (int) find_user(cvspasswd, tmpcvspasswd, user.name, find_action_remove);
    if (!result)
	error(1, 0, "User %s not found", user.name);
    return 1;
}

static char *global_tmppasswd = NULL;

static void
atexit_unlink_tmppasswd (void)
{
    unlink(global_tmppasswd);
    global_tmppasswd = NULL;
}

int 
passwd(argc, argv)
    int argc;
    char **argv;
{
    int c;
    enum passwd_action what_to_do = action_none;
    int need_password = 0;
    FILE *cvspasswd;
    FILE *tmpcvspasswd;
    char *filename, *tmpfilename, *password2;
    int len;
    char *p;
    int passwd_updated = 0;
    struct CVS_user user;

    if (argc == -1)
	usage (passwd_usage);
  
    user.name = NULL;
    user.password = NULL;
    user.old_password = NULL;
    user.system_user = NULL;
    user.email = NULL;
    user.gecos = NULL;
    user.status = NULL;

    optind = 0;
    while ((c = getopt (argc, argv, "a:e:d:x:c:u:i:t:m:w:o:pP:")) != -1)
    {
	switch (c) 
	{
	case 'a':
	    if (what_to_do != action_none)
		usage(passwd_usage);
	    what_to_do = action_add;
	    user.name = xstrdup (optarg);
	    user.status = xstrdup("Disabled");
	    need_password = 1;
	    break;
	  
	case 'e':
	    if (what_to_do != action_none)
		usage(passwd_usage);
	    what_to_do = action_enable;
	    user.name = xstrdup(optarg);
	    user.status = xstrdup("Enabled");
	    break;
	  
	case 'd':
	    if (what_to_do != action_none)
		usage(passwd_usage);
	    what_to_do = action_disable;
	    user.name = xstrdup(optarg);
	    user.status = xstrdup("Disabled");
	    break;
	  
	case 'x':
	    if (what_to_do != action_none)
		usage(passwd_usage);
	    what_to_do = action_remove;
	    user.name = xstrdup(optarg);
	    break;
	  
	case 'c':
	    if (what_to_do != action_none)
		usage(passwd_usage);
	    what_to_do = action_change;
	    user.name = xstrdup(optarg);
	    break;

	case 'u':
	    if ((what_to_do != action_add) && (what_to_do != action_change))
		usage(passwd_usage);
	    user.system_user = xstrdup(optarg);
	    break;

	case 'i':
	    if ((what_to_do != action_add) && (what_to_do != action_change))
		usage(passwd_usage);
	    user.gecos = xstrdup(optarg);
	    break;
	    
	case 't':
	    if ((what_to_do != action_add) && (what_to_do != action_change))
		usage(passwd_usage);
	    user.status = xstrdup(optarg);
	    break;

	case 'm':
	    if ((what_to_do != action_add) && (what_to_do != action_change))
		usage(passwd_usage);
	    user.email = xstrdup(optarg);
	    break;

	case 'p':
	    if (what_to_do == action_none) {
		user.name = xstrdup(current_parsed_root->username);
		need_password = 1;
		what_to_do = action_passwd;
	    } else {
		if (what_to_do != action_change)
		    usage(passwd_usage);
		need_password = 1;
	    }
	    break;

#ifdef SERVER_SUPPORT
	    /* username to change password to in server mode */
	case 'P':
	    if (server_active)
		if (what_to_do != action_passwd)
		    error(1, 0, "internal error: username sent by client in illegal mode");
		else
		    user.name = xstrdup(optarg);
	    else
		error(1, 0, "internal error: -P %s passed", optarg);
	    break;

	case 'w':
	    if (server_active)
		if (
		    (what_to_do != action_add) 
		    && (what_to_do != action_change)
		    && (what_to_do != action_passwd)
		    )
		    error(1, 0, "internal error: password sent by client in illegal mode");
		else
		    user.password = xstrdup(optarg);
	    else
		error(1, 0, "internal error: -w %s passed", optarg);
	    break;

	case 'o':
	    if (server_active)
		if (what_to_do != action_passwd) 
		    error(1, 0, "internal error: old password sent by client in illegal mode");
		else
		    user.old_password = xstrdup(optarg);
	    else
		error (1, 0, "internal error: -o %s passed", optarg);
	    break;
#endif
	  
	default:
	    error(1, 0, "invalid option -%c\n", c);
	    usage(passwd_usage);
	    break;
	}
      
    }

    if (what_to_do == action_none)
	usage(passwd_usage);

    if (!user.name)
	usage(passwd_usage);

    if (strcmp(user.name, "anonymous") == 0)
	error(1, 0, "anonymous users are not allowed to use cvs passwd\n");

    if (what_to_do == action_add)
	if (!user.system_user)
	    usage(passwd_usage);

    for (p = user.name; *p; p++)
	if (!(isalnum(*p) || (*p == '-')))
	    error(1, 0, "username \"%s\" may consist only of alphanumeric or dash characters\n", user.name);

    if (user.system_user)
	for (p = user.system_user; *p; p++)
	    if (!isalnum(*p) && !(*p == '.'))
		error(1, 0, "system user/group name \"%s\" may consist only " \
		      "of alphanumeric characters or \".\"", user.system_user);

    if (user.gecos)
	for (p = user.gecos; *p; p++)
	    if (*p == ':')
		error(1, 0, "GECOS field \"%s\" should not contain colons\n", user.gecos);

    if (user.email)
	for (p = user.email; *p; p++)
	    if (!isalnum(*p) && (!strchr("@.!-\"<>()", *p)))
		error(1, 0, "E-mail addresses may contain: \n\
\t - alphanumeric characters;\n\
\t - `@', `.', `!', `-', `\"', `<', `>, `(', `)'\n");

#ifdef CLIENT_SUPPORT
    if (current_parsed_root->isremote)
	start_server ();
#endif
	
#ifdef SERVER_SUPPORT
    if (!server_active)
#endif

	if (need_password)
	{
	    char *prompt_fmt = "Enter CVS password for user %s: ";
	    char *prompt2_fmt = "Enter CVS password for user %s again: ";
	    int len = strlen(prompt2_fmt) + strlen(user.name) + 1;
	    char *prompt = xmalloc(len);
	    
	    snprintf(prompt, len, prompt_fmt, user.name); 
	    user.password = xstrdup(getpass(prompt));

	    snprintf(prompt, len, prompt2_fmt, user.name);
	    password2 = getpass(prompt);
	    
	    if (strcmp(user.password, password2) != 0) 
		error(1, 0, "passwords do not match!\n");
	    free(prompt);
	}

#ifdef CLIENT_SUPPORT

    if (current_parsed_root->isremote)
    {
	switch (what_to_do)
	{
	case action_add: 
	    option_with_arg ("-a", user.name);
	    break;

	case action_change:
	    option_with_arg ("-c", user.name);
	    break;

	case action_passwd:
	    send_option_string ("-p");
	    option_with_arg ("-P", user.name);
	    break;

	case action_enable:
	    option_with_arg ("-e", user.name);
	    break;

	case action_disable:
	    option_with_arg ("-d", user.name);
	    break;

	case action_remove:
	    option_with_arg ("-x", user.name);
	    break;

	default:
	    error(1, 0, "internal error: invalid what_to_do\n");
	}

	if ((what_to_do == action_add) || (what_to_do == action_change))
	{
	    option_with_arg ("-u", user.system_user);

	    if (user.email)
		option_with_arg ("-m", user.email);

	    if (user.gecos)
		option_with_arg ("-t", user.gecos);

	    option_with_arg ("-w", user.password);
	}

	if (what_to_do == action_passwd) {
	    option_with_arg ("-o", descramble(get_cvs_password()));
	    option_with_arg ("-w", user.password);
	}

	send_to_server ("passwd\012", 0);

	return get_responses_and_close ();
    }
#endif

    if (what_to_do == action_passwd) {
	/* printf("Changing password for user %s (via cvspasswd)\n", user.name); */
	passwd_passwd(user);
	return 0;
    }

    len = strlen(current_parsed_root->directory) + 1 + 
	strlen(CVSROOTADM) + 1 +
	strlen(CVSROOTADM_CVSPASSWD) + 1;
    filename = xmalloc(len);
    snprintf(filename, len, "%s/%s/%s", 
	     current_parsed_root->directory, CVSROOTADM, CVSROOTADM_CVSPASSWD);
  
    cvspasswd = CVS_FOPEN(filename, "r");
    if ((cvspasswd == NULL) && !existence_error(errno))
	error(1, 0, "opening %s: %s", filename, strerror(errno));

#define TMP_SUFFIX ".tmp"

    tmpfilename = xmalloc(len + strlen(TMP_SUFFIX));
    strcpy(tmpfilename, filename);
    strcat(tmpfilename, TMP_SUFFIX);

    lockpasswd(current_parsed_root->directory);

    tmpcvspasswd = create_cvspasswd(tmpfilename);
    if (!tmpcvspasswd)
	error(1, 0, "opening %s: %s", tmpfilename, strerror(errno));
    global_tmppasswd = tmpfilename;
    atexit(atexit_unlink_tmppasswd);
  
    switch (what_to_do)
    {
    case action_add:
	passwd_updated = passwd_add_user(user, cvspasswd, tmpcvspasswd);
	break;

    case action_enable:
	/* printf("Enabling user %s\n", user.name); */
	passwd_updated = passwd_edit_user(user, cvspasswd, tmpcvspasswd);
	break;
      
    case action_disable:
	/* printf("Disabling user %s\n", user.name); */
	passwd_updated = passwd_edit_user(user, cvspasswd, tmpcvspasswd);
	break;

    case action_remove:
	/* printf("Removing user %s\n", user.name); */
	passwd_updated = passwd_remove_user(user, cvspasswd, tmpcvspasswd);
	break;

    case action_change:
	/* printf("Changing info about user %s\n", user.name); */
	passwd_updated = passwd_edit_user(user, cvspasswd, tmpcvspasswd);
	break;

    default:
	error(1, 0, "internal error: invalid what_to_do\n");
    }

    if (cvspasswd)
	fclose(cvspasswd);
    if (fclose(tmpcvspasswd) != 0)
	error(1, 0, "renaming %s to %s: %s", tmpfilename, filename, strerror (errno));
    if (passwd_updated)
	rename_file(tmpfilename, filename);
    else {
	unlink_file(tmpfilename);
	global_tmppasswd = NULL;
    }

    unlockpasswd(current_parsed_root->directory);

    return 0;
}

#endif /* SERVER_SUPPORT */
