/* 
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or (at
   your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   Copyright (c) Alexey Mahotkin <alexm@hsys.msk.ru> 2001

   "cvs racl" subcommand manages ACL directly to repository

*/

#include "config.h"

#include "cvs.h"
#include "acl.h"

#include "acl-admins.h"
#include "module-acl.h"
#include "repository.h"

#include <errno.h>
#include <stdlib.h>


static const char *const racl_usage[] = 
{
    "Usage: cvs racl [-r BRANCH] [grant|revoke] USER PERMISSION modules...\n",
    "\tPERMISSION could be one of those:\n",
    "\t\ttag\t\tAllow creating tags\n",
    "\t\ttag:PREFIX\tAllow tags whose names start with PREFIX\n",
    "\t\tbranch\t\tAllow creating branches\n",
    "\t\tbranch:PREFIX\tAllow branches whose names start with PREFIX\n",
    NULL
};

#ifdef SERVER_SUPPORT
static void
server_side_racl (int argc, char **argv)
{

    struct REPOSITORY *repos;
    char *subcommand;
    char *username;
    char *perm;
    char *prefix;
    int grant;

    if (!is_acl_admin(CVS_Username))
	error(1, 0, "%s is not allowed to manage ACLs", CVS_Username);

    argv++; /* skip 'cvs server' */
    argv++; /* skip 'racl' */
    
    repos = get_repository(current_parsed_root->directory);
    
    subcommand = *argv++;
    if (!subcommand)
	usage(racl_usage);
    
    if (strcmp(subcommand, "grant") == 0)
	grant = 1;
    else if (strcmp(subcommand, "revoke") == 0)
	grant = 0;
    else
	usage(racl_usage);
    
    username = *argv++;
    if (!username)
	usage(racl_usage);
    
    perm = *argv++;
    if (!perm)
	usage(racl_usage);
    
    if (strncmp(perm, "tag", strlen("tag")) == 0) {
	prefix = perm + strlen("tag") + 1;
    } else {
	error(1, 0, "Unknown permission: '%s'", perm);
    }
    
    while (*argv) {
	char *module = *argv;
	struct MODULE_ACL *mod_acl = get_module_acl(repos, module);
    
	if (grant) {
	    if (!grant_module_tagging(mod_acl, username, prefix))
		error(1, errno, "Error granting tag permission to %s for %s",
		      username, module);
        } else {
	    if (!revoke_module_tagging(mod_acl, username, prefix))
		error(1, errno, "Error revoking tag permission to %s for %s",
		      username, module);
        }
        argv++;
    }
    
    if (!save_module_acl(repos))
	error(1, errno, "Error saving module ACLS: %s", strerror(errno));
    
    free_repository(repos);
}
#endif

static int
client_side_racl (int argc, char **argv)
{
    int i;
    start_server();
    
    if (!supported_request ("racl"))
	error (1, 0, "server does not support racl");

    for (i = 0; i < argc; i++) {
	send_arg(argv[i]);
    }

    send_to_server ("racl\012", 0);

    return get_responses_and_close();
}


int
racl (int argc, char **argv)
{
    if (argc == -1)
	usage(racl_usage);
    
#ifdef SERVER_SUPPORT
    if (server_active)
	server_side_racl(argc, argv);
#endif

#ifdef CLIENT_SUPPORT
    if (current_parsed_root->isremote)
	return client_side_racl(argc, argv);
#endif

    return 0;
}
