/* 
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or (at
   your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   Copyright (c) Alexey Mahotkin <alexm@hsys.msk.ru> 2001
   Copyright (c) Alexey Morozov <morozov@novosoft.ru>, Novosoft, Inc. 2001

   Network client using rsh-style tunneling.

*/

#include "config.h"

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <string.h>
#include <errno.h>

#include "root.h"
#include "run.h"

#include "network-client.h"
#include "rsh-client.h"
#include "fd-client.h"

struct rsh_client_data {
    cvsroot_t *cvsroot;
    int rsh_pid;
    int toserver_fd;
    int fromserver_fd;
};

typedef struct rsh_client_data *RSHCD;


static int rsh_connect(NCLI);
static int rsh_disconnect(NCLI);

static int rsh_read(NCLI, char *buf, unsigned int len);
static int rsh_write(NCLI, char *buf, unsigned int len);
static int rsh_flush(NCLI);


struct network_client *
create_rsh_client (cvsroot_t *CVSROOT, ncli_error_cb error_cb)
{
    RSHCD rshcd;
    struct network_client *nc = malloc(sizeof(struct network_client));
    if (nc == NULL)
		return NULL;

    rshcd = malloc(sizeof(struct rsh_client_data));
    if (rshcd == NULL)
		return NULL;

    rshcd->cvsroot = CVSROOT;

    nc->connect = rsh_connect;
    nc->disconnect = rsh_disconnect;

    nc->read = rsh_read;
    nc->write = rsh_write;
    nc->flush = rsh_flush;

    nc->error = error_cb;

    nc->client_data = rshcd;

    return nc;
}


static int
rsh_connect (NCLI ncli)
{
    char *rsh = getenv("CVS_RSH");
    char *cvs = getenv("CVS_SERVER");
    char *argv[10]; /* eg: "rsh cvs.example.com -l vasya cvs server" */
    int i = 0;
    int rsh_pid;
    int toserver_fd, fromserver_fd;
    RSHCD rshcd = (RSHCD) ncli->client_data;

    if (rsh == NULL)
		rsh = "rsh";
    if (cvs == NULL)
		cvs = "cvs";

    argv[i++] = rsh;
    argv[i++] = rshcd->cvsroot->hostname;
    if (rshcd->cvsroot->username != NULL) {
		argv[i++] = "-l";
		argv[i++] = rshcd->cvsroot->username;
    }
    argv[i++] = cvs;
    argv[i++] = "server"; /* Buggy rsh clients are _not_ worked around here */
    argv[i++] = NULL;

    rsh_pid = piped_child(argv, &toserver_fd, &fromserver_fd);
	/* should we report errors somehow? */

    rshcd->rsh_pid = rsh_pid;
    rshcd->toserver_fd = toserver_fd;
    rshcd->fromserver_fd = fromserver_fd;
	return NCLI_SUCCESS;
}

static int
rsh_disconnect (ncli)
	NCLI ncli;
{
	RSHCD rshcd = (RSHCD) ncli->client_data;

	if (close (rshcd->toserver_fd) == -1)
	{
		ncli->error(ncli, "closing connection to %s: %s",
					rshcd->cvsroot->hostname, strerror(errno));
		return 0;
	};

	{
		int c;
		if (read(rshcd->fromserver_fd,(void*)&c,1) != 0)
		{
			ncli->error(ncli, "dying gasps from %s unexpected",
						rshcd->cvsroot->hostname);
			return 0;
		}
	}
	{
		int status; /* does anybody really need this? */
		if (rshcd->rsh_pid != -1 &&
			waitpid(rshcd->rsh_pid,&status,0) ==-1)
		{
			ncli->error(ncli, "waiting for process %d: %s",
						rshcd->rsh_pid, strerror(errno));
			return 0;
		}
	}
	rshcd->rsh_pid = -1;
	rshcd->toserver_fd = -1;
	rshcd->fromserver_fd = -1;
	return NCLI_SUCCESS;
}


static int
rsh_read (NCLI ncli, char *buf, unsigned int len)
{
    RSHCD rshcd = (RSHCD) ncli->client_data;
	int r = basic_fd_read(ncli,rshcd->fromserver_fd,buf,len);
	if (r == -1) {
		ncli->error(ncli,"Unable to read data from %s: %s",
					rshcd->cvsroot->hostname,strerror(errno));
		return -1;
	}
	return r;
}



static int
rsh_write (NCLI ncli, char *buf, unsigned int len)
{
    RSHCD rshcd = (RSHCD) ncli->client_data;
	int r = basic_fd_write(ncli,rshcd->toserver_fd,buf,len);
	if (r == NCLI_FAILURE) {
		ncli->error(ncli,"Unable to write data to %s: %s",
					rshcd->cvsroot->hostname,strerror(errno));
		return NCLI_FAILURE;
	}
	return NCLI_SUCCESS;
}


static int
rsh_flush (NCLI ncli)
{
    (void) ncli;

    return NCLI_SUCCESS;
}


